@@ -6,6 +6,7 @@ import nbformat
 import pytest
 import pkg_resources
 
+from copy import deepcopy
 from functools import lru_cache
 from textwrap import dedent
 from types import MethodType
@@ -118,66 +119,119 @@ def test_check_cm(capsys):
     _, stu = generate_impl()
     mfp = stu.values
 
-    with mock.patch("pybryt.student.tracing_on") as mocked_tracing, mock.patch("pybryt.student.tracing_off"):
-        check_cm = check(ref)
-        with check_cm:
-            check_cm._observed = mfp
-
-    captured = capsys.readouterr()
-    expected = dedent("""\
-        REFERENCE: foo
-        SATISFIED: True
-        MESSAGES:
-          - SUCCESS: Sorted the sample correctly
-          - SUCCESS: Computed the size of the sample
-          - SUCCESS: computed the correct median
-    """)
-    assert captured.out == expected
-
-    with mock.patch("pybryt.student.tracing_on") as mocked_tracing, mock.patch("pybryt.student.tracing_off"):
-        ref_filename = pkg_resources.resource_filename(__name__, os.path.join("files", "expected_ref.pkl"))
-        check_cm = check(ref_filename)
-        with check_cm:
-            check_cm._observed = mfp
-
-        check_cm2 = check([ref_filename])
-        assert check_cm._ref == check_cm2._ref
-
-    captured = capsys.readouterr()
-    expected = dedent("""\
-        REFERENCE: foo
-        SATISFIED: True
-        MESSAGES:
-          - SUCCESS: Sorted the sample correctly
-          - SUCCESS: Computed the size of the sample
-          - SUCCESS: computed the correct median
-    """)
-    assert captured.out == expected
-
-    # check no action when tracing
-    global __PYBRYT_TRACING__
-    __PYBRYT_TRACING__ = True
-
-    try:
+    with mock.patch.object(check, "_cache_check") as mocked_cache:
         with mock.patch("pybryt.student.tracing_on") as mocked_tracing, mock.patch("pybryt.student.tracing_off"):
+            check_cm = check(ref, cache=False)
+            with check_cm:
+                check_cm._observed = mfp
+
+            mocked_cache.assert_not_called()
+
+        captured = capsys.readouterr()
+        expected = dedent("""\
+            REFERENCE: foo
+            SATISFIED: True
+            MESSAGES:
+              - SUCCESS: Sorted the sample correctly
+              - SUCCESS: Computed the size of the sample
+              - SUCCESS: computed the correct median
+        """)
+        assert captured.out == expected
+
+        with mock.patch("pybryt.student.tracing_on") as mocked_tracing, mock.patch("pybryt.student.tracing_off"):
+            ref_filename = pkg_resources.resource_filename(__name__, os.path.join("files", "expected_ref.pkl"))
+            check_cm = check(ref_filename)
+            with check_cm:
+                check_cm._observed = mfp
+
+            mocked_cache.assert_called()
+
+            check_cm2 = check([ref_filename])
+            assert check_cm._ref == check_cm2._ref
+
+        captured = capsys.readouterr()
+        expected = dedent("""\
+            REFERENCE: foo
+            SATISFIED: True
+            MESSAGES:
+              - SUCCESS: Sorted the sample correctly
+              - SUCCESS: Computed the size of the sample
+              - SUCCESS: computed the correct median
+        """)
+        assert captured.out == expected
+
+        # check no action when tracing
+        global __PYBRYT_TRACING__
+        __PYBRYT_TRACING__ = True
+
+        try:
+            with mock.patch("pybryt.student.tracing_on") as mocked_tracing, mock.patch("pybryt.student.tracing_off"):
+                check_cm = check(ref)
+                with check_cm:
+                    pass
+
+                mocked_tracing.assert_not_called()
+
+        except:
+            __PYBRYT_TRACING__ = False
+            raise
+
+        else:
+            __PYBRYT_TRACING__ = False
+
+        # test errors
+        with pytest.raises(ValueError, match="Cannot check against an empty list of references"):
+            check([])
+
+        with pytest.raises(TypeError, match="Invalid values in the reference list"):
+            check([ref, "path", 1])
+
+    # check caching
+    with mock.patch("pybryt.student.tracing_on") as mocked_tracing, mock.patch("pybryt.student.tracing_off"):
+        with mock.patch("pybryt.student.StudentImplementation") as mocked_stu, \
+                mock.patch("pybryt.student.generate_report") as mocked_generate, \
+                mock.patch("pybryt.student.os.makedirs") as mocked_makedirs:
+            mocked_stu.from_footprint.return_value.check.return_value = [mock.MagicMock()]
+            mocked_stu.from_footprint.return_value.check.return_value[0].name = "foo"
+            # breakpoint()
             check_cm = check(ref)
             with check_cm:
-                pass
-            mocked_tracing.assert_not_called()
+                check_cm._observed = mfp
+            
+            mocked_makedirs.assert_called_with(".pybryt_cache", exist_ok=True)
+            mocked_stu.from_footprint.return_value.dump.assert_called()
+            mocked_stu.from_footprint.return_value.check.return_value[0].dump.assert_called_with(".pybryt_cache/foo_results.pkl")
 
-    except:
-        __PYBRYT_TRACING__ = False
-        raise
 
-    else:
-        __PYBRYT_TRACING__ = False
+def test_from_cache():
+    """
+    """
+    with mock.patch("pybryt.student.glob") as mocked_glob, \
+            mock.patch.object(StudentImplementation, "load") as mocked_load, \
+            mock.patch.object(StudentImplementation, "combine") as mocked_combine:
+        mocked_glob.return_value = [".pybryt_cache/student_impl_foo.pkl", ".pybryt_cache/student_impl_bar.pkl"]
+        StudentImplementation.from_cache(combine=False)
 
-    # test errors
-    with pytest.raises(ValueError, match="Cannot check against an empty list of references"):
-        check([])
+        mocked_load.assert_has_calls([mock.call(fp) for fp in mocked_glob.return_value])
+        mocked_combine.assert_not_called()
 
-    with pytest.raises(TypeError, match="Invalid values in the reference list"):
-        check([ref, "path", 1])
+        StudentImplementation.from_cache()
+
+        mocked_combine.assert_called()
+
+
+def test_combine():
+    """
+    """
+    _, stu = generate_impl()
+    stu2 = deepcopy(stu)
+    stu2.steps += 1
+    stu2.values.append(([1, 2, 3, 4], stu2.steps))
+
+    comb = StudentImplementation.combine([stu, stu2])
+    assert len(comb.values) == len(stu.values)  + 1
+    assert comb.steps == stu.steps + stu2.steps
+    assert comb.values[-1][1] == stu.steps + stu2.steps
 
 
 def test_generate_student_impls():
